@@ -83,7 +83,7 @@ E_VMEXT = objectify.ElementMaker(
     namespace=NSMAP['vmext'],
     nsmap={
         None: NSMAP['vcloud'],
-        'vmext': NSMAP['vmext']
+        'vmext': NSMAP['vmext'],
     })
 
 E_OVF = objectify.ElementMaker(
@@ -104,6 +104,9 @@ class ApiVersion(Enum):
     VERSION_31 = '31.0'
     VERSION_32 = '32.0'
     VERSION_33 = '33.0'
+    VERSION_34 = '34.0'
+    VERSION_35 = '35.0'
+    VERSION_36 = '36.0'
 
 
 # Important! Values must be listed in ascending order.
@@ -112,7 +115,10 @@ API_CURRENT_VERSIONS = [
     ApiVersion.VERSION_30.value,
     ApiVersion.VERSION_31.value,
     ApiVersion.VERSION_32.value,
-    ApiVersion.VERSION_33.value
+    ApiVersion.VERSION_33.value,
+    ApiVersion.VERSION_34.value,
+    ApiVersion.VERSION_35.value,
+    ApiVersion.VERSION_36.value
 ]
 
 
@@ -338,6 +344,7 @@ class EntityType(Enum):
         'application/vnd.vmware.vcloud.allocatedNetworkAddress+xml'
     AMQP_SETTINGS = 'application/vnd.vmware.admin.amqpSettings+xml'
     API_EXTENSIBILITY = 'application/vnd.vmware.vcloud.apiextensibility+xml'
+    API_FILTER = 'application/vnd.vmware.admin.apiFilter+xml'
     APPLICATION_BINARY = 'application/binary'
     CATALOG = 'application/vnd.vmware.vcloud.catalog+xml'
     CAPTURE_VAPP_PARAMS = \
@@ -371,12 +378,14 @@ class EntityType(Enum):
     EXTERNAL_NETWORK = 'application/vnd.vmware.admin.vmwexternalnet+xml'
     EXTERNAL_NETWORK_REFS = \
         'application/vnd.vmware.admin.vmwExternalNetworkReferences+xml'
-    INSTANTIATE_VAPP_TEMPLATE_PARAMS = \
-        'application/vnd.vmware.vcloud.instantiateVAppTemplateParams+xml'
     GUEST_CUSTOMIZATION_SECTION = \
         'application/vnd.vmware.vcloud.guestCustomizationSection+xml'
     HISTORIC_USAGE = \
         'application/vnd.vmware.vcloud.metrics.historicUsageSpec+xml'
+    HOST = 'application/vnd.vmware.admin.host+xml'
+    HOST_REFS = 'application/vnd.vmware.admin.vmwHostReferences+xml'
+    INSTANTIATE_VAPP_TEMPLATE_PARAMS = \
+        'application/vnd.vmware.vcloud.instantiateVAppTemplateParams+xml'
     JSON = 'application/json'
     LEASE_SETTINGS = 'application/vnd.vmware.vcloud.leaseSettingsSection+xml'
     MEDIA = 'application/vnd.vmware.vcloud.media+xml'
@@ -431,6 +440,8 @@ class EntityType(Enum):
     UNDEPLOY = 'application/vnd.vmware.vcloud.undeployVAppParams+xml'
     UPDATE_PROVIDER_VDC_STORAGE_PROFILES = \
         'application/vnd.vmware.admin.updateProviderVdcStorageProfiles+xml'
+    UPDATE_VDC_STORAGE_PROFILES = \
+        'application/vnd.vmware.admin.updateVdcStorageProfiles+xml'
     UPLOAD_VAPP_TEMPLATE_PARAMS = \
         'application/vnd.vmware.vcloud.uploadVAppTemplateParams+xml'
     USER = 'application/vnd.vmware.admin.user+xml'
@@ -441,6 +452,9 @@ class EntityType(Enum):
     VDC_COMPUTE_POLICY_REFERENCES = \
         "application/vnd.vmware.vcloud.vdcComputePolicyReferences+xml"
     VDC_REFERENCES = 'application/vnd.vmware.admin.vdcReferences+xml'
+    VDC_STORAGE_PROFILE = 'application/vnd.vmware.vcloud.vdcStorageProfile+xml'
+    VDC_STORAGE_PROFILE_ADMIN = \
+        'application/vnd.vmware.admin.vdcStorageProfile+xml'
     VDCS_PARAMS = 'application/vnd.vmware.admin.createVdcParams+xml'
     VIM_SERVER_REFS = 'application/vnd.vmware.admin.vmwVimServerReferences+xml'
     VIRTUAL_CENTER = 'application/vnd.vmware.admin.vmwvirtualcenter+xml'
@@ -749,26 +763,6 @@ class Client(object):
                  log_bodies=False):
         self._logger = None
         self._get_default_logger(file_name=log_file)
-        self._logger.setLevel(logging.DEBUG)
-        # This makes sure that we don't append a new handler to the logger
-        # every time we create a new client. Since all the client share the
-        # same underlying logger instance.
-        if not self._logger.handlers:
-            if log_file is not None:
-                # make sure the path to the log file is valid, create missing
-                # folders if required
-                Path(log_file).parent.mkdir(parents=True, exist_ok=True)
-                handler = logging.FileHandler(log_file)
-            else:
-                handler = logging.NullHandler()
-            formatter = logging.Formatter('%(asctime)-23.23s | '
-                                          '%(levelname)-5.5s | '
-                                          '%(name)-15.15s | '
-                                          '%(module)-15.15s | '
-                                          '%(funcName)-12.12s | '
-                                          '%(message)s')
-            handler.setFormatter(formatter)
-            self._logger.addHandler(handler)
 
         self._log_requests = log_requests
         self._log_headers = log_headers
@@ -819,15 +813,21 @@ class Client(object):
         if file_name is None:
             file_name = "vcd_pysdk.log"
         self._logger = logging.getLogger(file_name)
+        self._logger.setLevel(log_level)
         file = Path(file_name)
         if not file.exists():
             file.parent.mkdir(parents=True, exist_ok=True)
         if not self._logger.handlers:
-            default_log_handler = handlers.RotatingFileHandler(
+            log_handler = handlers.RotatingFileHandler(
                 filename=file_name, maxBytes=max_bytes,
                 backupCount=backup_count)
-            default_log_handler.setLevel(log_level)
-            self._logger.addHandler(default_log_handler)
+            formatter = logging.Formatter(
+                fmt='%(asctime)s | %(module)s:%(lineno)s - %(funcName)s '
+                    '| %(levelname)s :: %(message)s',
+                datefmt='%y-%m-%d %H:%M:%S')
+            log_handler.setFormatter(formatter)
+            log_handler.setLevel(log_level)
+            self._logger.addHandler(log_handler)
 
     def _negotiate_api_version(self):
         # If user provided API version we accept it, otherwise negotiate with
@@ -976,7 +976,10 @@ class Client(object):
             if sc != requests.codes.ok:
                 r = None
                 try:
-                    r = _objectify_response(response)
+                    if accept_type.lower() == 'application/json':
+                        r = response.json()
+                    else:
+                        r = _objectify_response(response)
                 except Exception:
                     pass
                 if r is not None:
@@ -1137,6 +1140,14 @@ class Client(object):
         """
         return self._vcloud_access_token
 
+    def should_verify_ssl(self):
+        """Retrieve boolean indicating if ssl verification is enabled.
+
+        :return: True if ssl verification is enabled, else False
+        :rtype: bool
+        """
+        return self._verify_ssl_certs
+
     def get_task_monitor(self):
         if self._task_monitor is None:
             self._task_monitor = _TaskMonitor(self)
@@ -1148,14 +1159,16 @@ class Client(object):
                     contents=None,
                     media_type=None,
                     objectify_results=True,
-                    params=None):
+                    params=None,
+                    extra_headers=None):
         response = self._do_request_prim(
             method,
             uri,
             self._session,
             contents=contents,
             media_type=media_type,
-            params=params)
+            params=params,
+            extra_headers=extra_headers)
 
         sc = response.status_code
         if sc in (requests.codes.ok,
@@ -1216,18 +1229,14 @@ class Client(object):
                 redacted_headers[key] = "[REDACTED]"
         return redacted_headers
 
-    def _log_request_response(self,
-                              response,
-                              request_body=None,
-                              skip_logging_response_body=False):
+    def _log_request_sent(self, method, uri, headers=[], request_body=None):
         if not self._log_requests:
             return
 
-        self._logger.debug('Request uri (%s): %s' % (response.request.method,
-                                                     response.request.url))
+        self._logger.debug(f"Request uri {method}: {uri}")
 
         if self._log_headers:
-            self._logger.debug(f"Request headers: {self._redact_headers(response.request.headers)}")  # noqa: E501
+            self._logger.debug(f"Request partial headers: {self._redact_headers(headers)}")  # noqa: E501
 
         if self._log_bodies and request_body is not None:
             if isinstance(request_body, str):
@@ -1236,6 +1245,15 @@ class Client(object):
                 body = request_body.decode(self.fsencoding)
             self._logger.debug('Request body: %s' % body)
 
+    def _log_request_response(self,
+                              response,
+                              skip_logging_response_body=False):
+        if not self._log_requests:
+            return
+
+        if self._log_headers:
+            self._logger.debug(f"Request full headers: {self._redact_headers(response.request.headers)}")  # noqa: E501
+
         self._logger.debug('Response status code: %s' % response.status_code)
 
         if self._log_headers:
@@ -1258,8 +1276,9 @@ class Client(object):
                          media_type=None,
                          accept_type=None,
                          auth=None,
-                         params=None):
-        headers = {}
+                         params=None,
+                         extra_headers=None):
+        headers = extra_headers or {}
         if media_type is not None:
             headers[self._HEADER_CONTENT_TYPE_NAME] = media_type
 
@@ -1279,6 +1298,9 @@ class Client(object):
             else:
                 data = etree.tostring(contents)
 
+        self._log_request_sent(
+            method=method, uri=uri, headers=headers, request_body=data)
+
         response = session.request(
             method,
             uri,
@@ -1288,7 +1310,7 @@ class Client(object):
             auth=auth,
             verify=self._verify_ssl_certs)
 
-        self._log_request_response(response=response, request_body=data)
+        self._log_request_response(response=response)
 
         return response
 
@@ -1304,6 +1326,7 @@ class Client(object):
         # retry efforts fail, we will fail the upload completely and return.
         for attempt in range(1, self._UPLOAD_FRAGMENT_MAX_RETRIES + 1):
             try:
+                self._log_request_sent(method='PUT', uri=uri, headers=headers)
                 response = self._session.put(
                     uri,
                     data=data,
@@ -1334,7 +1357,7 @@ class Client(object):
                           chunk_size=SIZE_1MB,
                           size=0,
                           callback=None):
-
+        self._log_request_sent(method='GET', uri=uri)
         response = self._session.get(
             uri, stream=True, verify=self._verify_ssl_certs)
         self._log_request_response(response, skip_logging_response_body=True)
@@ -1395,7 +1418,8 @@ class Client(object):
                       contents,
                       media_type,
                       params=None,
-                      objectify_results=True):
+                      objectify_results=True,
+                      extra_headers=None):
         """Posts to a resource link.
 
         Posts the specified contents to the specified resource. (Does an HTTP
@@ -1407,9 +1431,11 @@ class Client(object):
             contents=contents,
             media_type=media_type,
             objectify_results=objectify_results,
-            params=params)
+            params=params,
+            extra_headers=extra_headers)
 
-    def post_linked_resource(self, resource, rel, media_type, contents):
+    def post_linked_resource(self, resource, rel, media_type, contents,
+                             extra_headers=None):
         """Posts to a resource link.
 
         Posts the contents of the resource referenced by the link with the
@@ -1423,20 +1449,23 @@ class Client(object):
         try:
             return self.post_resource(
                 find_link(resource, rel, media_type).href, contents,
-                media_type)
+                media_type, extra_headers=extra_headers)
         except MissingLinkException as e:
             raise OperationNotSupportedException(
                 "Operation is not supported").with_traceback(e.__traceback__)
 
-    def get_resource(self, uri, params=None, objectify_results=True):
+    def get_resource(self, uri, params=None, objectify_results=True,
+                     extra_headers=None):
         """Gets the specified contents to the specified resource.
 
         This method does an HTTP GET.
         """
         return self._do_request(
-            'GET', uri, objectify_results=objectify_results, params=params)
+            'GET', uri, objectify_results=objectify_results, params=params,
+            extra_headers=extra_headers)
 
-    def get_linked_resource(self, resource, rel, media_type):
+    def get_linked_resource(self, resource, rel, media_type,
+                            extra_headers=None):
         """Gets the content of the resource link.
 
         Gets the contents of the resource referenced by the link with the
@@ -1451,16 +1480,20 @@ class Client(object):
             the link being not visible to the logged in user of the client.
         """
         try:
-            return self.get_resource(find_link(resource, rel, media_type).href)
+            return self.get_resource(find_link(resource, rel, media_type).href,
+                                     extra_headers=extra_headers)
         except MissingLinkException as e:
             raise OperationNotSupportedException(
                 "Operation is not supported").with_traceback(e.__traceback__)
 
-    def delete_resource(self, uri, params=None, force=False, recursive=False):
+    def delete_resource(self, uri, params=None, force=False, recursive=False,
+                        extra_headers=None):
         full_uri = '%s?force=%s&recursive=%s' % (uri, force, recursive)
-        return self._do_request('DELETE', full_uri, params=params)
+        return self._do_request('DELETE', full_uri, params=params,
+                                extra_headers=extra_headers)
 
-    def delete_linked_resource(self, resource, rel, media_type):
+    def delete_linked_resource(self, resource, rel, media_type,
+                               extra_headers=None):
         """Deletes the resource referenced by the link.
 
         Deletes the resource referenced by the link with the specified rel and
@@ -1471,7 +1504,8 @@ class Client(object):
         """
         try:
             return self.delete_resource(
-                find_link(resource, rel, media_type).href)
+                find_link(resource, rel, media_type).href,
+                extra_headers=extra_headers)
         except MissingLinkException as e:
             raise OperationNotSupportedException(
                 "Operation is not supported").with_traceback(e.__traceback__)
@@ -1585,6 +1619,7 @@ class Client(object):
     def get_typed_query(self,
                         query_type_name,
                         query_result_format=QueryResultFormat.REFERENCES,
+                        page=None,
                         page_size=None,
                         include_links=False,
                         qfilter=None,
@@ -1626,6 +1661,7 @@ class Client(object):
             query_type_name,
             self,
             query_result_format,
+            page=page,
             page_size=page_size,
             include_links=include_links,
             qfilter=qfilter,
@@ -1661,8 +1697,8 @@ class Client(object):
         Process the query link and fetch all vdc links and make a list of
             links.
 
-        :param lxml.objectify.ObjectifiedElement resource: the resource with
-            the links.
+        :param lxml.objectify.ObjectifiedElement resource: resource with
+            query links.
         :param RelationType rel: the rel of the desired link.
         :param str media_type: media type of content.
         :param str type: type of query.
@@ -1675,17 +1711,21 @@ class Client(object):
         for query_link in links:
             link_href = query_link.href
             list_of_links = []
-            if type in link_href.lower():
-                link_href = urllib.parse.unquote(link_href)
-                record_resource = self.get_resource(link_href)
-                for record in record_resource.getchildren():
-                    if record.tag != '{http://www.vmware.com/vcloud/v1.5}Link':
-                        link = E.Link()
-                        link.set('name', record.get('name'))
-                        link.set('href', record.get('href'))
-                        link.set('rel', rel.value)
-                        list_of_links.append(Link(link))
-                return list_of_links
+            if type not in link_href.lower():
+                continue
+            link_href = urllib.parse.unquote(link_href)
+            next_page_uri = link_href
+            while next_page_uri is not None:
+                record_resource = self.get_resource(next_page_uri)
+                next_page_uri = None
+                for record in record_resource.iterchildren():
+                    tag = etree.QName(record.tag)
+                    if tag.localname == 'Link':
+                        if record.get('rel') == RelationType.NEXT_PAGE.value:
+                            next_page_uri = record.get('href')
+                    else:
+                        list_of_links.append(Link(record))
+            return list_of_links
 
 
 def find_link(resource, rel, media_type, fail_if_absent=True, name=None):
@@ -1767,6 +1807,7 @@ class _AbstractQuery(object):
     def __init__(self,
                  query_result_format,
                  client,
+                 page=None,
                  page_size=None,
                  include_links=False,
                  qfilter=None,
@@ -1800,33 +1841,71 @@ class _AbstractQuery(object):
         self._query_result_format = query_result_format
         self._page_size = page_size
         self._include_links = include_links
+        self._query_all_pages = True
         self._page = 1
+        if page:
+            self._query_all_pages = False
+            self._page = page
 
-        self._filter = qfilter
-        if equality_filter is not None:
-            if self._filter is not None:
+        is_below_v35_query = float(client.get_api_version()) < float(ApiVersion.VERSION_35.value)  # noqa: E501
+        if is_below_v35_query:
+            self._filter = qfilter
+        else:
+            self._filter = self._escape_special_characters(qfilter)
+
+        if equality_filter:
+            if self._filter:
                 self._filter += ';'
             else:
                 self._filter = ''
             self._filter += equality_filter[0]
             self._filter += '=='
-            self._filter += equality_filter[1]
+            if is_below_v35_query:
+                self._filter += equality_filter[1]
+            else:
+                self._filter += self._escape_special_characters(
+                    urllib.parse.quote(equality_filter[1]))
 
         self._sort_desc = sort_desc
         self._sort_asc = sort_asc
 
         self.fields = fields
 
+    def _escape_special_characters(self, single_encoded_value_string):
+        """Escape vCD query specific special characters viz. ( ) ; ,.
+
+        This method accepts a url encoded string and replaces encoded
+        instances of ( ) ; , with their escaped version.
+
+        :param str single_encoded_value_string: string to be processed
+
+        :return: string with vCD query specific special characters escaped
+
+        :rtype: str
+        """
+        val = single_encoded_value_string
+        if val:
+            val = val.replace('%28', '%5C%28')
+            val = val.replace('%29', '%5C%29')
+            val = val.replace('%2C', '%5C%2C')
+            val = val.replace('%3B', '%5C%3B')
+        return val
+
     def execute(self):
         """Executes query and returns results.
 
-        :return: A generator to returns results.
+        :return: If a specific page number is set, return a dictionary which
+            contains results of the specific page or else,
+            return a generator to return results
 
-        :rtype: generator object
+        :rtype: dict if page is set else a generator object
         """
+        # create query href
         query_href = self._find_query_uri(self._query_result_format)
         if query_href is None:
             raise OperationNotSupportedException('Unable to execute query.')
+
+        # build query uri
         query_uri = self._build_query_uri(
             query_href,
             self._page,
@@ -1834,7 +1913,41 @@ class _AbstractQuery(object):
             self._filter,
             self._include_links,
             fields=self.fields)
-        return self._iterator(self._client.get_resource(query_uri))
+
+        if self._query_all_pages:
+            # Iterate over all the pages present to return all the resources
+            return self._iterator(self._client.get_resource(query_uri))
+
+        # return the resources in the present in the required page number
+        result = {}
+        query_results = self._client.get_resource(query_uri)
+        result['resultTotal'] = int(query_results.get('total'))
+        result['nextPageUri'] = None
+        if self._page * self._page_size < result['resultTotal']:
+            # build query uri for the next page
+            result['nextPageUri'] = self._build_query_uri(
+                query_href,
+                self._page + 1,
+                self._page_size,
+                self._filter,
+                self._include_links,
+                fields=self.fields)
+        if self._page > 1:
+            # build query uri for the previous page
+            result['previousPageUri'] = self._build_query_uri(
+                query_href,
+                self._page - 1,
+                self._page_size,
+                self._filter,
+                self._include_links,
+                fields=self.fields)
+        resources = []
+        for r in query_results.iterchildren():
+            tag = etree.QName(r.tag)
+            if tag.localname != 'Link':
+                resources.append(r)
+        result['values'] = resources
+        return result
 
     def _iterator(self, query_results):
         while True:
@@ -1889,14 +2002,20 @@ class _AbstractQuery(object):
             uri += '&pageSize='
             uri += str(page_size)
 
-        if qfilter is not None:
-            # filterEncoded=true directs vCD to decode the individual filter
-            # values, i.e. the value after each ==.
-            uri += '&filterEncoded=true&filter='
-            # Need to encode the value of filter param again to escape special
-            # characters like ',', ';' which have special meaning in context of
-            # query filter.
-            uri += urllib.parse.quote(qfilter)
+        if qfilter:
+            if float(self._client.get_api_version()) < float(ApiVersion.VERSION_35.value):  # noqa: E501
+                # filterEncoded=true directs vCD to decode the individual
+                # filter values, i.e. the value after each ==.
+                uri += '&filterEncoded=true&filter='
+                # Need to encode the value of filter param again to escape
+                # special characters like ',', ';' which have special meaning
+                # in context of query filter.
+                uri += urllib.parse.quote(qfilter)
+            else:
+                # api v35.0 (vCD Zeus) onwards vCD doesn't accept double
+                # encoded filter values, neither the filterEncoded query param
+                # has any effect
+                uri += f"&filter={qfilter}"
 
         if fields is not None:
             uri += '&fields='
@@ -1918,6 +2037,7 @@ class _TypedQuery(_AbstractQuery):
                  query_type_name,
                  client,
                  query_result_format,
+                 page=None,
                  page_size=None,
                  include_links=False,
                  qfilter=None,
@@ -1928,6 +2048,7 @@ class _TypedQuery(_AbstractQuery):
         super(_TypedQuery, self).__init__(
             query_result_format,
             client,
+            page=page,
             page_size=page_size,
             include_links=include_links,
             qfilter=qfilter,
