@@ -46,21 +46,24 @@ from pyvcloud.vcd.exceptions import OperationNotSupportedException
 from pyvcloud.vcd.exceptions import UploadException
 from pyvcloud.vcd.metadata import Metadata
 from pyvcloud.vcd.system import System
+from pyvcloud.vcd.utils import extract_id
 from pyvcloud.vcd.utils import get_admin_href
 from pyvcloud.vcd.utils import get_non_admin_href
 from pyvcloud.vcd.utils import get_safe_members_in_tar_file
 from pyvcloud.vcd.utils import is_admin
 from pyvcloud.vcd.utils import retrieve_compute_policy_id_from_href
 from pyvcloud.vcd.utils import to_dict
+from pyvcloud.vcd.utils import update_vm_compute_policy_element
+from pyvcloud.vcd.utils import VDC_COMPUTE_POLICY_MAX_API_VERSION
+from pyvcloud.vcd.utils import VDC_COMPUTE_POLICY_MIN_API_VERSION
+from pyvcloud.vcd.utils import VM_SIZING_POLICY_MIN_API_VERSION
 
 # Uptil pyvcloud v20.0.0 1 MB was the default chunk size,
 # in constrast vCD H5 UI uses 50MB for upload chunk size,
 # 10MB is a happy medium between 50MB and 1MB.
 DEFAULT_CHUNK_SIZE = 10 * 1024 * 1024
 
-VDC_COMPUTE_POLICY_MIN_API_VERSION = float(ApiVersion.VERSION_32.value)
-VDC_COMPUTE_POLICY_MAX_API_VERSION = float(ApiVersion.VERSION_33.value)
-VM_SIZING_POLICY_MIN_API_VERSION = float(ApiVersion.VERSION_33.value)
+TENANT_CONTEXT_HDR = 'X-VMWARE-VCLOUD-TENANT-CONTEXT'
 
 
 class Org(object):
@@ -118,9 +121,11 @@ class Org(object):
         if self.resource is None:
             self.reload()
         payload = E.AdminCatalog(E.Description(description), name=name)
+        extra_headers = {TENANT_CONTEXT_HDR:
+                         extract_id(self.resource.attrib['id'])}
         return self.client.post_linked_resource(
             self.resource, RelationType.ADD, EntityType.ADMIN_CATALOG.value,
-            payload)
+            payload, extra_headers=extra_headers)
 
     def delete_catalog(self, name):
         """Delete a catalog in the organization.
@@ -938,18 +943,20 @@ class Org(object):
         :param str email: email address of the user.
         :param str telephone: The telephone of the user.
         :param str im: instant message address of the user.
-        :param str alert_email: email address where alerts should be sent.
+        :param str alert_email: email address where alerts should be sent
+            (Deprecated).
         :param str alert_email_prefix: string to prepend to alert message
-            subject line.
+            subject line (Deprecated).
         :param str stored_vm_quota: quota of vApps that this user can store.
         :param str deployed_vm_quota: quota of vApps that this user can deploy
             concurrently.
         :param bool is_group_role: indicates if the user has a group role.
-        :param bool is_default_cached: indicates if user should be cached.
+        :param bool is_default_cached: indicates if user should be cached
+            (Deprecated).
         :param bool is_external: indicates if user is imported from an external
             source.
         :param bool is_alert_enabled: if, True will enable email alert for the
-            user.
+            user (Deprecated).
         :param bool is_enabled: if True, will enable the user after creation
             else will let the user remain disabled after creation.
 
@@ -966,11 +973,7 @@ class Org(object):
             E.Telephone(telephone),
             E.IsEnabled(is_enabled),
             E.IM(im),
-            E.IsAlertEnabled(is_alert_enabled),
-            E.AlertEmailPrefix(alert_email_prefix),
-            E.AlertEmail(alert_email),
             E.IsExternal(is_external),
-            E.IsDefaultCached(is_default_cached),
             E.IsGroupRole(is_group_role),
             E.StoredVmQuota(stored_vm_quota),
             E.DeployedVmQuota(deployed_vm_quota),
@@ -980,7 +983,8 @@ class Org(object):
         return self.client.post_linked_resource(
             resource_admin, RelationType.ADD, EntityType.USER.value, user)
 
-    def update_user(self, user_name, is_enabled=None, role_name=None):
+    def update_user(self, user_name, is_enabled=None, role_name=None,
+                    password=None):
         """Update an user.
 
         :param str user_name: username of the user.
@@ -994,17 +998,26 @@ class Org(object):
         :rtype: lxml.objectify.ObjectifiedElement
         """
         user = self.get_user(user_name)
-        if is_enabled is not None:
-            if hasattr(user, 'IsEnabled'):
-                user['IsEnabled'] = E.IsEnabled(is_enabled)
-        if role_name:
-            if hasattr(user, 'Role'):
-                role = self.get_role_record(role_name)
-                user['Role'] = E.Role(href=role.get('href'))
+        updated_user = E.User(name=user_name)
 
-        if is_enabled or role_name:
+        if is_enabled is not None:
+            updated_user['IsEnabled'] = E.IsEnabled(is_enabled)
+            user['IsEnabled'] = updated_user['IsEnabled']
+
+        if role_name is not None:
+            role = self.get_role_record(role_name)
+            updated_user['Role'] = E.Role(href=role.get('href'))
+            user['Role'] = updated_user['Role']
+        else:
+            updated_user["Role"] = user["Role"]
+
+        if password is not None:
+            updated_user["Password"] = E.Password(password)
+            user["Password"] = updated_user["Password"]
+
+        if is_enabled or role_name or password:
             return self.client.put_resource(
-                user.get('href'), user, EntityType.USER.value)
+                user.get('href'), updated_user, EntityType.USER.value)
 
         return user
 
@@ -1044,7 +1057,7 @@ class Org(object):
         if self.client.is_sysadmin():
             resource_type = ResourceType.ADMIN_USER.value
             org_filter = 'org==%s' % \
-                urllib.parse.quote_plus(self.resource.get('href'))
+                urllib.parse.quote(self.resource.get('href'))
         query = self.client.get_typed_query(
             resource_type,
             query_result_format=QueryResultFormat.RECORDS,
@@ -1153,7 +1166,7 @@ class Org(object):
         if self.client.is_sysadmin():
             resource_type = ResourceType.ADMIN_ROLE.value
             org_filter = 'org==%s' % \
-                urllib.parse.quote_plus(self.resource.get('href'))
+                urllib.parse.quote(self.resource.get('href'))
 
         query = self.client.get_typed_query(
             resource_type,
@@ -1578,6 +1591,106 @@ class Org(object):
             resource_admin, RelationType.ADD, EntityType.VDCS_PARAMS.value,
             params)
 
+    def update_org_vdc(self,
+                       vdc_name,
+                       description=None,
+                       allocation_model=None,
+                       cpu_units=None,
+                       cpu_allocated=None,
+                       cpu_limit=None,
+                       mem_units=None,
+                       mem_allocated=None,
+                       mem_limit=None,
+                       nic_quota=None,
+                       network_quota=None,
+                       vm_quota=None,
+                       resource_guaranteed_memory=None,
+                       resource_guaranteed_cpu=None,
+                       vcpu_in_mhz=None,
+                       is_thin_provision=None,
+                       is_enabled=None):
+        """Update an Organization VDC in the current organization.
+
+        :param str vdc_name: name of the org vdc.
+        :param str description: new description of the org vdc.
+        :param str allocation_model: updated allocation model used
+            by the vdc. Accepted values are 'AllocationVApp',
+            'AllocationPool' or 'ReservationPool'.
+        :param str cpu_units: updated unit for compute capacity allocated
+            to the vdc. Accepted values are 'MHz' or 'GHz'.
+        :param int cpu_allocated: updated capacity that is committed to be
+            available.
+        :param int cpu_limit: updated capacity limit relative to the value
+            specified for allocation.
+        :param str mem_units: updated unit for memory capacity allocated to
+            the vdc. Acceptable values are 'MB' or 'GB'.
+        :param int mem_allocated: updated memory capacity that is committed
+            to be available.
+        :param int mem_limit: updated memory capacity limit relative to the
+            value specified for allocation.
+        :param int nic_quota: updated maximum number of virtual NICs allowed
+            in the vdc.
+        :param int network_quota: updated maximum number of network objects
+            that can be deployed in the vdc.
+        :param int vm_quota: updated maximum number of VMs that can be created
+            in the vdc.
+        :param float resource_guaranteed_memory: updated percentage of
+            allocated CPU resources guaranteed to vApps deployed in
+            the vdc.
+        :param float resource_guaranteed_cpu: updated percentage of allocated
+            memory resources guaranteed to vApps deployed in the vdc.
+        :param int vcpu_in_mhz: updated clock frequency, in MegaHertz,
+            for any virtual CPU that is allocated to a VM.
+        :param bool is_thin_provision: True to request thin provisioning.
+        :param bool is_enabled: True / False  if the vdc is enabled for use
+            by the organization users.
+
+        :return: a task to update the vdc.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        vdc = self.get_vdc(vdc_name, is_admin_operation=True)
+        if vdc is None:
+            raise EntityNotFoundException("{0} is not found".format(vdc_name))
+
+        if description:
+            vdc['Description'] = E.Description(description)
+        if allocation_model:
+            vdc['AllocationModel'] = E.AllocationModel(allocation_model)
+        if cpu_units:
+            vdc.ComputeCapacity.Cpu.Units = E.Units(cpu_units)
+        if cpu_allocated:
+            vdc.ComputeCapacity.Cpu.Allocated = E.Allocated(cpu_allocated)
+        if cpu_limit:
+            vdc.ComputeCapacity.Cpu.Limit = E.Limit(cpu_limit)
+        if mem_units:
+            vdc.ComputeCapacity.Memory.Units = E.Units(mem_units)
+        if mem_allocated:
+            vdc.ComputeCapacity.Memory.Allocated = E.Allocated(mem_allocated)
+        if mem_limit:
+            vdc.ComputeCapacity.Memory.Limit = E.Limit(mem_limit)
+        if nic_quota:
+            vdc['NicQuota'] = E.NicQuota(nic_quota)
+        if network_quota:
+            vdc['NetworkQuota'] = E.NetworkQuota(network_quota)
+        if vm_quota:
+            vdc['VmQuota'] = E.VmQuota(vm_quota)
+        if resource_guaranteed_memory:
+            vdc['ResourceGuaranteedCpu'] = E.ResourceGuaranteedCpu(
+                resource_guaranteed_cpu)
+        if resource_guaranteed_cpu:
+            vdc['ResourceGuaranteedMemory'] = E.ResourceGuaranteedMemory(
+                resource_guaranteed_memory)
+        if vcpu_in_mhz:
+            vdc['VCpuInMhz'] = E.VCpuInMhz(vcpu_in_mhz)
+        if is_thin_provision:
+            vdc['IsThinProvision'] = E.IsThinProvision(is_thin_provision)
+        if is_enabled is not None:
+            vdc['IsEnabled'] = E.IsEnabled(is_enabled)
+
+        return self.client.put_resource(
+            vdc.get("href"), vdc, EntityType.VDC_ADMIN.value)
+
     def get_vdc(self, name, is_admin_operation=False):
         """Retrieves resource of an org vdc identified by its name.
 
@@ -1817,17 +1930,126 @@ class Org(object):
         catalog_item = self.get_catalog_item(catalog_name, catalog_item_name)
         return catalog_item.Entity.get('href')
 
+    def assign_placement_policy_to_vapp_template_vms(self,
+                                                     catalog_name,
+                                                     catalog_item_name,
+                                                     placement_policy_href,
+                                                     placement_policy_final=False):  # noqa: E501
+        """Assign placement policy to all vms of a given vApp template.
+
+        The placement policy is identified by its href.
+
+        :param str catalog_name: Name of the catalog that contains the template
+        :param str catalog_item_name: Name of the template (catalog item)
+        :param str placement_policy_href: href of the placement policy
+        :param bool placement_policy_final: if set to True, the placement
+            policy can't be overridden
+
+        :return: an object of type EntityType.TASK XML which represents
+                the asynchronous task that is updating virtual application
+                template. If the given vApp doesn't have any vm in it, then
+                None will be returned.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+
+        :raises: OperationNotSupportedException: if the api version is not
+            supported.
+        """
+        api_version = float(self.client.get_api_version())
+        if api_version < VM_SIZING_POLICY_MIN_API_VERSION:
+            raise OperationNotSupportedException(
+                f"Unsupported API version. Received '{api_version}' but "
+                f"atleast'{VM_SIZING_POLICY_MIN_API_VERSION}' is required.")
+
+        template_resource_href = self.get_vapp_template_href(
+            catalog_name, catalog_item_name)
+        template_resource = self.client.get_resource(template_resource_href)
+        template_update_required = False
+        if hasattr(template_resource, 'Children') and \
+                hasattr(template_resource.Children, 'Vm'):
+            vms = template_resource.Children.Vm
+            for vm in vms:
+                template_update_required = \
+                    update_vm_compute_policy_element(api_version,
+                                                     vm,
+                                                     placement_policy_href=placement_policy_href,  # noqa: E501
+                                                     placement_policy_final=placement_policy_final)  # noqa: E501
+
+            if template_update_required:
+                return self.client.put_resource(
+                    template_resource_href,
+                    template_resource,
+                    media_type=EntityType.VAPP_TEMPLATE.value)
+
+    def assign_sizing_policy_to_vapp_template_vms(self,
+                                                  catalog_name,
+                                                  catalog_item_name,
+                                                  sizing_policy_href,
+                                                  sizing_policy_final=False):
+        """Assign sizing policy to all vms in a given vApp template.
+
+        The sizing policy is identified by its href.
+
+        :param str catalog_name: Name of the catalog that contains the template
+        :param str catalog_item_name: Name of the template (catalog item)
+        :param str sizing_policy_href: href of the sizing policy
+        :param bool sizing_policy_final: If set to True, the sizing policy
+            can't be overridden
+
+        :return: an object of type EntityType.TASK XML which represents
+            the asynchronous task that is updating virtual application
+            template. If the given vApp doesn't have any vm in it, then None
+            will be returned.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+
+        :raises: OperationNotSupportedException: if the api version is not
+            supported.
+        """
+        api_version = float(self.client.get_api_version())
+        if api_version < VM_SIZING_POLICY_MIN_API_VERSION:
+            raise OperationNotSupportedException(
+                f"Unsupported API version. Received '{api_version}' but "
+                f"atleast'{VM_SIZING_POLICY_MIN_API_VERSION}' is required.")
+
+        template_resource_href = self.get_vapp_template_href(
+            catalog_name, catalog_item_name)
+        template_resource = self.client.get_resource(template_resource_href)
+        template_update_required = False
+        if hasattr(template_resource, 'Children') and \
+                hasattr(template_resource.Children, 'Vm'):
+            vms = template_resource.Children.Vm
+            for vm in vms:
+                template_update_required = \
+                    update_vm_compute_policy_element(api_version,
+                                                     vm,
+                                                     sizing_policy_href=sizing_policy_href,  # noqa: E501
+                                                     sizing_policy_final=sizing_policy_final)  # noqa: E501
+
+            if template_update_required:
+                return self.client.put_resource(
+                    template_resource_href,
+                    template_resource,
+                    media_type=EntityType.VAPP_TEMPLATE.value)
+
     def assign_compute_policy_to_vapp_template_vms(self,
                                                    catalog_name,
                                                    catalog_item_name,
-                                                   compute_policy_href):
+                                                   compute_policy_href,
+                                                   sizing_policy_final=False):
         """Assign compute policy to all vms in a given vApp template.
 
+        NOTE: This function is deprecated as VdcComputePolicy Tags are
+        deprecated. When used with api_version >= 33.0, the function tries to
+        add a sizing policy with href 'compute_policy_href' to template VMs
+
         The compute policy is identified by its href.
 
         :param str catalog_name: Name of the catalog that contains the template
         :param str catalog_item_name: Name of the template (catalog item)
         :param str compute_policy_href: href of the compute policy
+        :param bool sizing_policy_final: If set to True, the compute policy
+            can't be overridden.
 
         :return: an object of type EntityType.TASK XML which represents
             the asynchronous task that is updating virtual application
@@ -1849,6 +2071,7 @@ class Org(object):
         template_resource_href = self.get_vapp_template_href(
             catalog_name, catalog_item_name)
         template_resource = self.client.get_resource(template_resource_href)
+        template_update_required = False
 
         if hasattr(template_resource, 'Children') and \
                 hasattr(template_resource.Children, 'Vm'):
@@ -1858,8 +2081,11 @@ class Org(object):
 
                 if api_version <= VDC_COMPUTE_POLICY_MAX_API_VERSION:
                     if hasattr(vm, 'VdcComputePolicy'):
+                        if vm.VdcComputePolicy.get('href', '') != compute_policy_href:  # noqa: E501
+                            template_update_required = True
                         vdc_compute_policy_element = vm.VdcComputePolicy
                     else:
+                        template_update_required = True
                         vdc_compute_policy_element = E.VdcComputePolicy()
                         date_created_node.addprevious(vdc_compute_policy_element)  # noqa: E501
 
@@ -1868,36 +2094,25 @@ class Org(object):
                     vdc_compute_policy_element.set('type', 'application/json')
 
                 if api_version >= VM_SIZING_POLICY_MIN_API_VERSION:
-                    if hasattr(vm, 'ComputePolicy'):
-                        compute_policy_element = vm.ComputePolicy
-                    else:
-                        compute_policy_element = E.ComputePolicy()
-                        date_created_node.addprevious(compute_policy_element)
+                    template_update_required = \
+                        update_vm_compute_policy_element(api_version,
+                                                         vm,
+                                                         sizing_policy_href=compute_policy_href,  # noqa: E501
+                                                         sizing_policy_final=sizing_policy_final)  # noqa: E501
 
-                    if hasattr(compute_policy_element, 'VmSizingPolicy'):
-                        vm_sizing_policy_element = compute_policy_element.VmSizingPolicy  # noqa: E501
-                    else:
-                        vm_sizing_policy_element = E.VmSizingPolicy()
-                        compute_policy_element.append(vm_sizing_policy_element)
-                        compute_policy_element.append(
-                            E.VmSizingPolicyFinal('false'))
-
-                    vm_sizing_policy_element.set('href', compute_policy_href)
-                    vm_sizing_policy_element.set('id', policy_id)
-                    vm_sizing_policy_element.set('type', 'application/json')
-
-            return self.client.put_resource(
-                template_resource_href,
-                template_resource,
-                media_type=EntityType.VAPP_TEMPLATE.value)
+            if template_update_required:
+                return self.client.put_resource(
+                    template_resource_href,
+                    template_resource,
+                    media_type=EntityType.VAPP_TEMPLATE.value)
 
     def remove_compute_policy_from_vapp_template_vms(
             self, catalog_name, catalog_item_name, compute_policy_href=None):
         """Remove compute policy from all vms in a given vApp template.
 
         The compute policy is identified by its href. If compute policy is not
-        provided, all vdc compute policies (i.e. vm sizing policies) will be
-        removed.
+        provided, all vdc compute policies (i.e. vm sizing policies and
+        vm placement policies) will be removed.
 
         :param str catalog_name: Name of the catalog that contains the template
         :param str catalog_item_name: Name of the template (catalog item)
@@ -1938,16 +2153,27 @@ class Org(object):
                     if not policy_id or policy_id == vm_compute_policy_id:
                         vm.remove(vm.VdcComputePolicy)
                         template_update_required = True
-                if hasattr(vm, 'ComputePolicy') and hasattr(vm.ComputePolicy, 'VmSizingPolicy'):  # noqa: E501
-                    vm_compute_policy_id = \
-                        vm.ComputePolicy.VmSizingPolicy.get('id')
-                    if not policy_id or policy_id == vm_compute_policy_id:
-                        vm.ComputePolicy.remove(
-                            vm.ComputePolicy.VmSizingPolicy)
-                        template_update_required = True
-                        if hasattr(vm.ComputePolicy, 'VmSizingPolicyFinal'):  # noqa: E501
+                if hasattr(vm, 'ComputePolicy'):
+                    if hasattr(vm.ComputePolicy, 'VmSizingPolicy'):
+                        vm_sizing_policy_id = \
+                            vm.ComputePolicy.VmSizingPolicy.get('id')
+                        if not policy_id or policy_id == vm_sizing_policy_id:
                             vm.ComputePolicy.remove(
-                                vm.ComputePolicy.VmSizingPolicyFinal)
+                                vm.ComputePolicy.VmSizingPolicy)
+                            template_update_required = True
+                            if hasattr(vm.ComputePolicy, 'VmSizingPolicyFinal'):  # noqa: E501
+                                vm.ComputePolicy.remove(
+                                    vm.ComputePolicy.VmSizingPolicyFinal)
+                    if hasattr(vm.ComputePolicy, 'VmPlacementPolicy'):
+                        vm_placement_policy_id = \
+                            vm.ComputePolicy.VmPlacementPolicy.get('id')
+                        if not policy_id or policy_id == vm_placement_policy_id:  # noqa: E501
+                            vm.ComputePolicy.remove(
+                                vm.ComputePolicy.VmPlacementPolicy)
+                            template_update_required = True
+                            if hasattr(vm.ComputePolicy, 'VmPlacementPolicyFinal'):  # noqa: E501
+                                vm.ComputePolicy.remove(
+                                    vm.ComputePolicy.VmPlacementPolicyFinal)
 
         if template_update_required:
             return self.client.put_resource(
