@@ -34,6 +34,7 @@ from pyvcloud.vcd.exceptions import MultipleRecordsException
 from pyvcloud.vcd.exceptions import OperationNotSupportedException
 from pyvcloud.vcd.metadata import Metadata
 from pyvcloud.vcd.utils import retrieve_compute_policy_id_from_href
+from pyvcloud.vcd.utils import update_vm_compute_policy_element
 from pyvcloud.vcd.utils import uri_to_api_uri
 
 VDC_COMPUTE_POLICY_MIN_API_VERSION = float(ApiVersion.VERSION_32.value)
@@ -102,6 +103,20 @@ class VM(object):
                     return record.VimServerRef.get('name')
         return None
 
+    def get_moid(self):
+        """Returns the vCenter MoRef of this vm.
+
+        :return: MoRef of this vm.
+
+        :rtype: str
+        """
+        if hasattr(self.get_resource(), 'VCloudExtension'):
+            return self.get_resource().VCloudExtension[
+                '{' + NSMAP['vmext'] + '}VmVimInfo'][
+                    '{' + NSMAP['vmext'] + '}VmVimObjectRef'][
+                        '{' + NSMAP['vmext'] + '}MoRef'].text
+        return None
+
     def get_cpus(self):
         """Returns the number of CPUs in the vm.
 
@@ -129,19 +144,54 @@ class VM(object):
         return int(
             self.resource.VmSpecSection.MemoryResourceMb.Configured.text)
 
-    def update_compute_policy(self, compute_policy_href, compute_policy_id=None):  # noqa: E501
+    def remove_placement_policy(self):
+        """Remove placement policy associated with the VM.
+
+        :return: an object containing EntityType.TASK XML data which represents
+            the asynchronous task that updates the vm.
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        api_version = float(self.client.get_api_version())
+        if api_version < VM_SIZING_POLICY_MIN_API_VERSION:
+            raise OperationNotSupportedException(
+                f"Unsupported API version. Received '{api_version}' but "
+                f"'{VM_SIZING_POLICY_MIN_API_VERSION}' is required.")
+        vm_resource = self.get_resource()
+        if not hasattr(vm_resource.ComputePolicy, 'VmSizingPolicy'):
+            raise ValueError('Unable to remove placement policy if '
+                             ' Sizing Policy is not present')
+
+        # Remove VdcComputePolicy from request as we know that we are dealing
+        # with API version >= 33.0
+        if hasattr(vm_resource, 'VdcComputePolicy'):
+            vm_resource.remove(vm_resource.VdcComputePolicy)
+
+        vm_resource.ComputePolicy.remove(
+            vm_resource.ComputePolicy.VmPlacementPolicy)
+        if hasattr(vm_resource.ComputePolicy, 'VmPlacementPolicyFinal'):
+            vm_resource.ComputePolicy.remove(
+                vm_resource.ComputePolicy.VmPlacementPolicyFinal)
+
+        reconfigure_vm_link = find_link(self.resource,
+                                        RelationType.RECONFIGURE_VM,
+                                        EntityType.VM.value)
+        return self.client.post_resource(reconfigure_vm_link.href,
+                                         vm_resource,
+                                         EntityType.VM.value)
+
+    def update_compute_policy(self, compute_policy_href=None,
+                              placement_policy_href=None):
         """Updates the Compute Policy of this VM.
 
         For api v32, the VdcComputePolicy element is modified. For api v33 and
         above, VmSizingPolicy is modified.
 
         :param str compute_policy_href: href of the target compute policy
-        :param str compute_policy_id: id of the target compute policy. This is
-            a redundant param, keeping it, to preserve compatibility with
-            older clients using this method.
+            (also used to update sizing policy for api version >=33.0)
+        :param str placement_policy_href: target placement policy href.
 
         :return: an object containing EntityType.TASK XML data which represents
-            the asynchronous task that updates the vm.
+            the asynchronous task that updates the vm if update is needed.
 
         :rtype: lxml.objectify.ObjectifiedElement
         """
@@ -150,29 +200,53 @@ class VM(object):
             raise OperationNotSupportedException(
                 f"Unsupported API version. Received '{api_version}' but "
                 f"'{VDC_COMPUTE_POLICY_MIN_API_VERSION}' is required.")
-
-        if not compute_policy_id:
-            compute_policy_id = \
-                retrieve_compute_policy_id_from_href(compute_policy_href)
+        if api_version < VM_SIZING_POLICY_MIN_API_VERSION and \
+                placement_policy_href:
+            raise OperationNotSupportedException(
+                f"Unsupported API version. Received '{api_version}' but "
+                f"'{VM_SIZING_POLICY_MIN_API_VERSION}' is required.")
+        if not compute_policy_href and not placement_policy_href:
+            return
 
         vm_resource = self.get_resource()
+        # To check if update is required
+        is_update_required = False
+        date_created_node = vm_resource.find('{http://www.vmware.com/vcloud/v1.5}DateCreated')  # noqa: E501
 
-        if api_version <= VDC_COMPUTE_POLICY_MAX_API_VERSION:
-            vm_resource.VdcComputePolicy.set('href', compute_policy_href)
-            vm_resource.VdcComputePolicy.set('id', compute_policy_id)
+        if api_version < VDC_COMPUTE_POLICY_MAX_API_VERSION:
+            if compute_policy_href:
+                compute_policy_id = \
+                    retrieve_compute_policy_id_from_href(compute_policy_href)
+            if compute_policy_href:
+                if hasattr(vm_resource, 'VdcComputePolicy'):
+                    if vm_resource.VdcComputePolicy.get('href', '') != compute_policy_href:  # noqa: E501
+                        is_update_required = True
+                    vdc_compute_policy_element = vm_resource.VdcComputePolicy
+                else:
+                    is_update_required = True
+                    vdc_compute_policy_element = E.VdcComputePolicy()
+                    date_created_node.addprevious(vdc_compute_policy_element)
+                vdc_compute_policy_element.set('href', compute_policy_href)
+                vdc_compute_policy_element.set('id', compute_policy_id)
+                vdc_compute_policy_element.set('type', 'application/json')
+        else:
+            # No need to send VdcComputePolicy in request if >= 33.0
+            if hasattr(vm_resource, 'VdcComputePolicy'):
+                vm_resource.remove(vm_resource.VdcComputePolicy)
 
-        if api_version >= VM_SIZING_POLICY_MIN_API_VERSION:
-            vm_resource.ComputePolicy.VmSizingPolicy.set('href',
-                                                         compute_policy_href)
-            vm_resource.ComputePolicy.VmSizingPolicy.set('id',
-                                                         compute_policy_id)
+            is_update_required = \
+                update_vm_compute_policy_element(api_version,
+                                                 vm_resource,
+                                                 sizing_policy_href=compute_policy_href,  # noqa: E501
+                                                 placement_policy_href=placement_policy_href)  # noqa: E501
 
-        reconfigure_vm_link = find_link(self.resource,
-                                        RelationType.RECONFIGURE_VM,
-                                        EntityType.VM.value)
-        return self.client.post_resource(reconfigure_vm_link.href,
-                                         vm_resource,
-                                         EntityType.VM.value)
+        if is_update_required:
+            reconfigure_vm_link = find_link(self.resource,
+                                            RelationType.RECONFIGURE_VM,
+                                            EntityType.VM.value)
+            return self.client.post_resource(reconfigure_vm_link.href,
+                                             vm_resource,
+                                             EntityType.VM.value)
 
     def modify_cpu(self, virtual_quantity, cores_per_socket=None):
         """Updates the number of CPUs of a vm.
@@ -375,6 +449,75 @@ class VM(object):
         return self._perform_power_operation(
             rel=RelationType.POWER_RESET, operation_name='power reset')
 
+    def suspend(self):
+        """Suspend the vm.
+
+        :return: an object containing EntityType.TASK XML data which represents
+            the asynchronous task that is suspending the VM.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        self.get_resource()
+        return self._perform_power_operation(
+            rel=RelationType.POWER_SUSPEND, operation_name='suspend')
+
+    def discard_suspended_state(self):
+        """Discard the suspended state of the vm.
+
+        :return: an object containing EntityType.TASK XML data which represents
+            the asynchronous task that is discarding the vm's suspended state.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        self.get_resource()
+        return self.client.post_linked_resource(
+            self.resource, RelationType.DISCARD_SUSPENDED_STATE, None, None)
+
+    def edit_name(self, name):
+        """Edit name of the vm.
+
+        :param str name: New name of the vm. It is mandatory.
+
+        :return: an object containing EntityType.TASK XML data which represents
+            the asynchronous task that is reconfiguring the vm.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        if name is None or name.isspace():
+            raise InvalidParameterException("Name can't be None or empty")
+        self.get_resource()
+        params = E.Vm(
+            name=name.strip()
+        )
+        return self.client.post_linked_resource(
+            self.resource, RelationType.RECONFIGURE_VM,
+            EntityType.VM.value, params)
+
+    def edit_hostname(self, hostname):
+        """Edit hostname (computer name) of the vm.
+
+        :param str hostname: new hostname of the vm. It is mandatory.
+
+        :return: an object containing EntityType.TASK XML data which represents
+            the asynchronous task that is updating the vm's guest customization
+            section.
+
+        :rtype: lxml.objectify.ObjectifiedElement
+        """
+        if hostname is None or hostname.isspace():
+            raise InvalidParameterException("Hostname can't be None or empty")
+        self.get_resource()
+        # Get current values and overwrite only ComputerName
+        from lxml import objectify
+        params = self.resource.GuestCustomizationSection
+        params.ComputerName = objectify.DataElement(
+            hostname.strip(), nsmap='', _pytype='')
+
+        return self.client.put_resource(
+            self.resource.GuestCustomizationSection.get('href'),
+            params,
+            EntityType.GUEST_CUSTOMIZATION_SECTION.value)
+
     def deploy(self, power_on=True, force_customization=False):
         """Deploys the vm.
 
@@ -525,30 +668,33 @@ class VM(object):
 
         :rtype: list
         """
-        nics = []
-        self.get_resource()
-        if hasattr(self.resource.NetworkConnectionSection,
-                   'PrimaryNetworkConnectionIndex'):
-            primary_index = self.resource.NetworkConnectionSection.\
-                PrimaryNetworkConnectionIndex.text
+        # get network connection section.
+        net_conn_section = self.get_resource().NetworkConnectionSection
 
-        for nc in self.resource.NetworkConnectionSection.NetworkConnection:
-            nic = {}
-            nic[VmNicProperties.INDEX.value] = nc.NetworkConnectionIndex.text
-            nic[VmNicProperties.CONNECTED.value] = nc.IsConnected.text
-            nic[VmNicProperties.PRIMARY.value] = (
-                primary_index == nc.NetworkConnectionIndex.text)
-            nic[VmNicProperties.ADAPTER_TYPE.
-                value] = nc.NetworkAdapterType.text
-            nic[VmNicProperties.NETWORK.value] = nc.get(
-                VmNicProperties.NETWORK.value)
-            nic[VmNicProperties.IP_ADDRESS_MODE.
-                value] = nc.IpAddressAllocationMode.text
-            if hasattr(nc, 'IpAddress'):
-                nic[VmNicProperties.IP_ADDRESS.value] = nc.IpAddress.text
-            if hasattr(nc, 'MACAddress'):
-                nic[VmNicProperties.MAC_ADDRESS.value] = nc.MACAddress.text
-            nics.append(nic)
+        nics = []
+        if hasattr(net_conn_section, 'PrimaryNetworkConnectionIndex'):
+            primary_index = net_conn_section.PrimaryNetworkConnectionIndex.text
+            self.primary_index = primary_index
+
+        if hasattr(net_conn_section, 'NetworkConnection'):
+            for nc in net_conn_section.NetworkConnection:
+                nic = {}
+                nic[VmNicProperties.INDEX.value] = \
+                    int(nc.NetworkConnectionIndex.text)
+                nic[VmNicProperties.CONNECTED.value] = nc.IsConnected.text
+                nic[VmNicProperties.PRIMARY.value] = (
+                    primary_index == nc.NetworkConnectionIndex.text)
+                nic[VmNicProperties.ADAPTER_TYPE.
+                    value] = nc.NetworkAdapterType.text
+                nic[VmNicProperties.NETWORK.value] = nc.get(
+                    VmNicProperties.NETWORK.value)
+                nic[VmNicProperties.IP_ADDRESS_MODE.
+                    value] = nc.IpAddressAllocationMode.text
+                if hasattr(nc, 'IpAddress'):
+                    nic[VmNicProperties.IP_ADDRESS.value] = nc.IpAddress.text
+                if hasattr(nc, 'MACAddress'):
+                    nic[VmNicProperties.MAC_ADDRESS.value] = nc.MACAddress.text
+                nics.append(nic)
         return nics
 
     def delete_nic(self, index):
@@ -592,31 +738,6 @@ class VM(object):
             net_conn_section, RelationType.EDIT,
             EntityType.NETWORK_CONNECTION_SECTION.value, net_conn_section)
 
-    def suspend(self):
-        """Suspend the vm.
-
-        :return: an object containing EntityType.TASK XML data which represents
-            the asynchronous task that is suspending the VM.
-
-        :rtype: lxml.objectify.ObjectifiedElement
-        """
-        self.get_resource()
-        return self._perform_power_operation(
-            rel=RelationType.POWER_SUSPEND, operation_name='suspend')
-
-    def discard_suspended_state(self):
-        """Discard suspended state of the vm.
-
-        :return: an object containing EntityType.TASK XML data which represents
-                    the asynchronous task that is discarding suspended state
-                    of VM.
-
-        :rtype: lxml.objectify.ObjectifiedElement
-        """
-        self.get_resource()
-        return self.client.post_linked_resource(
-            self.resource, RelationType.DISCARD_SUSPENDED_STATE, None, None)
-
     def install_vmware_tools(self):
         """Install vmware tools in the vm.
 
@@ -931,7 +1052,6 @@ class VM(object):
             self.resource.BootOptions.BootDelay.addnext(ebs)
         if storage_policy_href is not None:
             storage_policy_res = self.client.get_resource(storage_policy_href)
-            print(storage_policy_href)
             self.resource.StorageProfile.set('href', storage_policy_href)
             self.resource.StorageProfile.set('id',
                                              storage_policy_res.get('id'))
@@ -1041,10 +1161,10 @@ class VM(object):
             for disk in disk_list.Item:
                 if disk['{' + NSMAP['rasd'] + '}Description'] == 'Hard disk':
                     vhs_disk_info = {
-                        'diskElementName': disk[
-                            '{' + NSMAP['rasd'] + '}ElementName'],
-                        'diskVirtualQuantityInBytes': disk[
-                            '{' + NSMAP['rasd'] + '}VirtualQuantity']
+                        'diskElementName': str(disk[
+                            '{' + NSMAP['rasd'] + '}ElementName']),
+                        'diskVirtualQuantityInBytes': int(disk[
+                            '{' + NSMAP['rasd'] + '}VirtualQuantity'])
                     }
                     result.append(vhs_disk_info)
 
@@ -1236,7 +1356,7 @@ class VM(object):
                                  EntityType.RELOCATE_PARAMS.value,
                                  relocate_params)
 
-    def update_nic(self, network_name,
+    def update_nic(self, network_name, nic_id=0,
                    is_connected=False,
                    is_primary=False,
                    ip_address_mode=None,
@@ -1263,25 +1383,27 @@ class VM(object):
         nic_found = False
         for network in net_conn_section.NetworkConnection:
             if network.get('network') == network_name:
-                nic_found = True
-                if ip_address is not None:
-                    network.IpAddress = E.IpAddress(ip_address)
-                network.IsConnected = E.IsConnected(is_connected)
-                if ip_address_mode is not None:
-                    network.IpAddressAllocationMode = \
-                        E.IpAddressAllocationMode(ip_address_mode)
-                if adapter_type is not None:
-                    network.NetworkAdapterType = E.NetworkAdapterType(
-                        adapter_type)
-                if is_primary:
-                    nic_index = network.NetworkConnectionIndex
-                break
+                if network.NetworkConnectionIndex == nic_id:
+                    nic_found = True
+                    if ip_address is not None:
+                        network.IpAddress = E.IpAddress(ip_address)
+                    network.IsConnected = E.IsConnected(is_connected)
+                    if ip_address_mode is not None:
+                        network.IpAddressAllocationMode = \
+                            E.IpAddressAllocationMode(ip_address_mode)
+                    if adapter_type is not None:
+                        network.NetworkAdapterType = E.NetworkAdapterType(
+                            adapter_type)
+                    if is_primary:
+                        nic_index = network.NetworkConnectionIndex
+                    break
 
         if nic_found is False:
             raise EntityNotFoundException(
                 'Vapp with name \'%s\' not found.' % network_name)
 
         if is_primary:
+            nic_index = int(nic_index.text)
             net_conn_section.PrimaryNetworkConnectionIndex = \
                 E.PrimaryNetworkConnectionIndex(nic_index)
 
@@ -1528,6 +1650,27 @@ class VM(object):
             put_resource(uri, vm_capabilities_section,
                          EntityType.VM_CAPABILITIES_SECTION.value)
 
+    def get_vm_extra_config_elements(self):
+        """Get vmw:ExtraConfig section of VM.
+
+        :return: list of all ExtraConfig elements
+        :rtype: list[lxml.objectify.StringElement]
+        """
+        vm_resource = self.get_resource()
+        return vm_resource.xpath(
+            'ovf:VirtualHardwareSection/vmw:ExtraConfig',
+            namespaces=NSMAP)
+
+    def list_vm_extra_config_info(self):
+        """List VM extra config key-value pairs.
+
+        :return: dictionary which contains VM ExtraConfig info
+        :rtype: dict
+        """
+        extra_config_items = self.get_vm_extra_config_elements()
+        return {item.get(f"{{{NSMAP['vmw']}}}key"): item.get(
+            f"{{{NSMAP['vmw']}}}value") for item in extra_config_items}
+
     def get_boot_options(self):
         """Get boot options of VM.
 
